DEFAULT_CAT = '_DEFAULT_'

#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

# GLOBAL execution parameters
RESAMPLING_RATE = 'resampling_rate_in_secs'
TRIP_GAP_THRESHOLD = 'trip_gap_threshold_in_secs'
TIMESTAMP_COLUMN = 'timestamp_column'
CATEGORY_COLUMN = 'category_column'

# PRETRANSFORM specific
FEATURES = 'features'
EVAL_FEATURES = 'eval_features'
LABELS = 'labels'
RESAMPLING = 'resampling'
FILLING = 'filling'
LABEL_TUNING = 'label_tuning'

DEFAULT_PREFIX = '_default'

# SEQUENCEGEN specific
SEQUENCE_LENGTH = 'sequence_length'
SEQUENCE_SHIFT = 'sequence_shift'
